""" Functions for Arch Linux functionality """
from subprocess import check_call, check_output, CalledProcessError
from .logger import get_logger

log = get_logger('arch') # pylint: disable=invalid-name

def pkgbuild_arrstr_to_list(s):
    """ Remove any redundant whitespace, quotes, and return a list """
    s = s.replace("'", '')
    s = s.replace('"', '')
    s = s.replace('\t', ' ')
    s = s.replace('\n', ' ')
    while '  ' in s:
        s = s.replace('  ', ' ')
    return s.split(' ')

def get_installed_packages():
    """ Leverage pacman to get installed packages """

    package_strings = check_output([
        'pacman',
        '-Q',
    ],
                                   universal_newlines=True)

    installed = []
    for pkg in package_strings.split('\n'):
        installed.append(pkg.split(' ')[0])
    return installed

def check_official_repository(pkgname):
    """ Check for an official package """
    log.info("checking the official repository for {}".format(pkgname))
    try:
        cmd = [
            'pacman',
            '-Ssq',
            '^{}$'.format(pkgname),
        ]
        log.debug("Running: {}".format(' '.join(cmd)))
        check_call(cmd)
        return True
    except CalledProcessError:
        return False

def install_official_package(pkgname, noconfirm=False):
    """ Use pacman to install a package from the official repos """

    cmd = [
        'sudo',
        'pacman',
        '-S',
    ]

    if noconfirm:
        cmd.append('--noconfirm')

    cmd.append(pkgname)

    log.debug("Running: {}".format(' '.join(cmd)))

    check_call(cmd)

    return [pkgname]



#!/bin/env python
""" Download and install packages from AUR """
import sys
import argparse
from .logger import get_logger, set_debug_logging
from .const import (
    WARNING_BANNER,
    BUILD_DIR,
    EXIT_CANCELLED,
)
from .package import pkgname_to_package
from .aur import install

log = get_logger('cli') # pylint: disable=invalid-name

def main():
    """ CLI entry point function """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('pkgname', metavar='PKGNAME', type=str, nargs='+',
                        help='Packages to install from AUR')
    parser.add_argument('-i', '--install', action='store_true', default=False,
                        help="Install the package, don't just download and unpack")
    parser.add_argument('-d', '--dir', type=str, default=None,
                        help="Use this directory to build")
    parser.add_argument('-y', '--noconfirm', action='store_true', default=False,
                        help="Do not confirm before installing packages, and do not review PKGBUILDs. NOT RECOMMENDED")
    parser.add_argument('--debug', action='store_true', default=False,
                        help="Show debug messages")

    args = parser.parse_args()

    if args.debug:
        set_debug_logging()

    if args.install and not args.noconfirm:

        for ln in WARNING_BANNER:
            print(ln)

        confirm = input().lower()

        if confirm != 'y':
            log.warning("User cancelled", file=sys.stderr)
            sys.exit(EXIT_CANCELLED)

    builddir = args.dir or BUILD_DIR
    for pkg in args.pkgname:
        installed = install(pkgname_to_package(pkg),
                            skip_install=(not args.install),
                            builddir=builddir,
                            noconfirm=args.noconfirm)

    log.info("Sucsessfully installed/built these pakcages: {}".format(' '.join(installed)))

if __name__ == '__main__':
    main()



""" Console logging functionality """
import logging
import coloredlogs

logger = logging.getLogger('sink') # pylint: disable=invalid-name

coloredlogs.install(level=logging.INFO,
                    fmt='%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s')

def get_logger(name):
    """ Return a child logger for use within a specific module """
    log = logger.getChild(name)
    return log

def set_debug_logging():
    """ Set logging output to DEBUG """
    logger.setLevel(logging.DEBUG)



""" Sink - Arch Linux AUR Package Installer """
__version__ = '0.1.0b2'
__author__ = 'Mike Shultz'
__copyright__ = "Copyright 2018 Mike Shultz"
__license__ = 'GPLv3'
__maintainer__ = __author__
__email__ = 'mike@mikeshultz.com'
__status__ = 'Alpha'



""" Various defs of things like constants and exceptions """
from datetime import datetime
from pathlib import Path

# URL Templates
AUR_TMPL_URL = "https://aur.archlinux.org/packages/{}/"
AUR_PKGBUILD_URL = "https://aur.archlinux.org/cgit/aur.git/plain/PKGBUILD?h={}"
SNAPSHOT_TMPL_URL = "https://aur.archlinux.org/cgit/aur.git/snapshot/{}.tar.gz"

# Paths
TMP_DIR = Path("/tmp/.sink-{}".format(datetime.now().isoformat().replace(':', '').replace('.', '')))
BUILD_DIR = TMP_DIR.joinpath("build")

WARNING_BANNER = [
    "========================THIS IS A DANGEROUS ACTION========================",
    "AUR Packages are user-generated and can not be trusted. It's also a good",
    "idea to make sure your system is up to date first as some packages may be",
    "looking for the most up to date libraries when building."
    "",
    "Are you sure you want to proceed? [y,N]",
]

# Exit codes
EXIT_OK = 0
EXIT_CANCELLED = 2
EXIT_ERROR = 3

# Regex for yo butts
PKGBUILD_DEPS_REGEX = r'depends\=\(([\w\d\s\-\>\<\=\.\n\'\"]+)\)'
PKGBUILD_MAKEDEPS_REGEX = r'makedepends\=\(([\w\d\s\-\>\<\=\.\n\'\"]+)\)'
PKGBUILD_DEP_W_VERSION = r'^([\w\d\-]+)([\>\<\=\d\-\.]*)$'

# Possible useful answers to a yes/no prompt
ANSWERS_POSITIVE = ['y', 'ye', 'yes']
ANSWERS_NEGATIVE = ['n', 'no', 'negatory']

# Exceptions
class SinkException(Exception): pass # pylint: disable=missing-docstring,multiple-statements
class PackageNotFound(SinkException): pass # pylint: disable=missing-docstring,multiple-statements



""" Package object and related """
import re
import semantic_version
from .const import PKGBUILD_DEP_W_VERSION

# pylint: disable=too-few-public-methods
class Package:
    """ Package object that handles all the weirdness

        ref: https://wiki.archlinux.org/index.php/PKGBUILD#depends
    """
    def __init__(self, pkgname):
        self.raw = pkgname
        self.spec = None
        self.pkgname = self._parse_pkgname(pkgname)

    def __repr__(self):
        return self.raw

    def _parse_pkgname(self, pkgname):
        """ Parse the given package name """

        # Not allowed in PKGBUILD
        assert not ('>' in pkgname and '<' in pkgname), "PKGBUILD dep format is invalid for {}".format(pkgname)

        parsed = re.match(PKGBUILD_DEP_W_VERSION, pkgname)

        # Assume it's just a package name if it doesn't parse out
        if not parsed or not parsed.group(2):
            self.pkgname = pkgname
        else:
            self.pkgname = parsed.group(1)
            # Set the version spec if we need to
            self.spec = semantic_version.Spec(parsed.group(2))

        return self.pkgname


def pkgname_to_package(pkgname):
    """ Convert a plain pkgname string to a Package object """
    if isinstance(pkgname, Package):
        return pkgname
    return Package(pkgname)



""" Functions to interact with the AUR """
import sys
import re
from pathlib import Path
from subprocess import check_call, CalledProcessError

import requests

from .const import (
    BUILD_DIR,
    SNAPSHOT_TMPL_URL,
    AUR_PKGBUILD_URL,
    PKGBUILD_DEPS_REGEX,
    PKGBUILD_MAKEDEPS_REGEX,
    EXIT_OK,
    EXIT_ERROR,
    PackageNotFound,
    SinkException,
    ANSWERS_POSITIVE,
    ANSWERS_NEGATIVE,
)
from .package import pkgname_to_package
from .arch import (
    get_installed_packages,
    check_official_repository,
    install_official_package,
    pkgbuild_arrstr_to_list,
)
from .logger import get_logger

log = get_logger('aur') # pylint: disable=invalid-name

def prompt_yes_no(msg):
    """ Prompt a user for a yes/no question """
    while True:
        inp = input(msg).lower()
        if inp in ANSWERS_POSITIVE:
            return True
        elif inp in ANSWERS_NEGATIVE:
            return False

def unpack_snapshot(sfile, pkgname):
    """ Unpack the snapshot """

    cmd = [
        'tar',
        '-xzf',
        str(sfile.resolve()),
        '-C',
        str(BUILD_DIR.resolve())
    ]
    
    log.debug("Running: {}".format(' '.join(cmd)))

    check_call(cmd)

    return BUILD_DIR.joinpath(pkgname)

def makepkg(pkgdir, noconfirm=False):
    """ Run makepkg """

    log.debug("Building package in {}".format(pkgdir))

    cmd = ['makepkg']
    if noconfirm:
        cmd.append('--noconfirm')

    log.debug("Running: {}".format(' '.join(cmd)))

    return check_call(cmd, cwd=pkgdir)

def makepkg_install(pkgdir, noconfirm=False):
    """ Run makepkg """

    log.info("Installing package from {}".format(pkgdir))

    cmd = ['makepkg', '-i']
    if noconfirm:
        cmd.append('--noconfirm')

    log.debug("Running: {}".format(' '.join(cmd)))

    return check_call(cmd, cwd=pkgdir)

def download(pkgname):
    """ Download an AUR snapshot for pkgname """

    log.info("Downloading {}...".format(SNAPSHOT_TMPL_URL.format(pkgname)))

    download_dest = BUILD_DIR.joinpath("{}.tar.gz".format(pkgname))
    cmd = [
        'curl',
        '-s',
        '--output',
        str(download_dest),
        SNAPSHOT_TMPL_URL.format(pkgname),
    ]

    log.debug("Running: {}".format(' '.join(cmd)))

    check_call(cmd)
    return download_dest

def fetch_pkgbuild(pkgname, dep_of=None):
    """ Retreive a PKGBUILD for a specific package """

    resp = requests.get(AUR_PKGBUILD_URL.format(pkgname))
    if resp.status_code == 404:
        msg = "Package {} does not exist"
        if dep_of is not None:
            msg += " (dependency of {})"
            msg = msg.format(pkgname, dep_of)
        else:
            msg = msg.format(pkgname)
        raise PackageNotFound(msg)
    elif resp.status_code != 200:
        raise SinkException("Unknown error from AUR.  Received code {}".format(resp.code))
    else:
        assert resp.status_code == 200

    return resp.content.decode('utf-8')

def fetch_deps(pkgname, dep_of=None):
    """ Get the deps from a pkgbuild """
    pkgbuild = fetch_pkgbuild(pkgname, dep_of)

    deps = []
    make_deps = []
    deps_match = re.search(PKGBUILD_DEPS_REGEX, pkgbuild)
    makedeps_match = re.search(PKGBUILD_MAKEDEPS_REGEX, pkgbuild)

    if not deps_match and not makedeps_match:
        log.debug("Unable to find either depends or makedepends in PKGBUILD")
    else:
        if deps_match:
            deps = pkgbuild_arrstr_to_list(deps_match.group(1))
        if makedeps_match:
            make_deps = pkgbuild_arrstr_to_list(makedeps_match.group(1))
        if len(deps) + len(make_deps) > 0:
            deps = sorted(list(set(deps + make_deps)))

    if len(deps) > 0:
        map(pkgname_to_package, deps)

    return deps

def confirm_pkgbuild(pkgdir):
    """ Display a PKGBUILD from a given dir for the user to review """

    CONFIRM_SUFFIX = '.confirm'
    pkgbuild = pkgdir.joinpath('PKGBUILD')

    if not pkgbuild.is_file():
        raise Exception("PKGBUILD missing")

    # Load PKGBUILD into vim
    try:
        check_call(['vim', '{}'.format(pkgbuild)])
    except CalledProcessError:
        log.debug("vim exited non-zero.")
        return False

    if prompt_yes_no("Continue building PKGBUILD? [y/n] "):
        return True

    return False

def install(pkgname, dep_of=None, skip_install=False, builddir=BUILD_DIR, noconfirm=False):
    """ Install a package from AUR """

    installed = []

    pkg = pkgname_to_package(pkgname)
    if not isinstance(builddir, Path):
        builddir = Path(builddir)

    log.debug("Preparing to install {}...".format(pkgname))

    # Make sure this package isn't already installed
    installed_packages = get_installed_packages()
    if pkg.pkgname in installed_packages:
        log.info("{} already installed.".format(pkgname))
        return [pkg.pkgname]

    # Check the official repos
    has_official_package = check_official_repository(pkg.pkgname)

    # Install the official package, or pretend if skipped
    if has_official_package:
        if skip_install:
            return [pkg.pkgname]
        try:
            return install_official_package(pkg.raw, noconfirm=noconfirm)
        except CalledProcessError:
            log.exception("Error installing package {} from official repository.".format(pkg.pkgname))
            sys.exit(EXIT_ERROR)

    # Dependencies first!
    deps = fetch_deps(pkg.pkgname, dep_of)
    if len(deps) > 0:
        for d in deps:
            installed.extend(install(d,
                                     dep_of=pkg.pkgname,
                                     skip_install=skip_install,
                                     noconfirm=noconfirm))

    # Make sure the temporary directory exists
    builddir.mkdir(mode=0o700, parents=True, exist_ok=True)

    # Download the PKGBUILD
    dest_file = None
    try:
        dest_file = download(pkg.pkgname)
    except CalledProcessError:
        log.exception("Error downloading snapshot for {}".format(pkg.pkgname))
        sys.exit(EXIT_ERROR)

    if not dest_file or not dest_file.exists():
        log.error("Unexpected file missing for {}".format(pkg.pkgname))
        sys.exit(EXIT_ERROR)

    # Extract
    dest_dir = None
    try:
        dest_dir = unpack_snapshot(dest_file, pkg.pkgname)
    except CalledProcessError:
        log.exception("Error extracting snapshot {}".format(dest_file))
        sys.exit(EXIT_ERROR)

    if not dest_dir or not dest_dir.exists():
        log.error("Extracted package missing for {}".format(pkg.pkgname))
        sys.exit(EXIT_ERROR)

    # Confirm the PKGBUILD with the user
    if not noconfirm:
        user_confirmed = confirm_pkgbuild(dest_dir)
        if not user_confirmed:
            log.warn("User declined PKGBUILD for {}".format(pkg.pkgname))
            sys.exit(EXIT_OK)

    # Build
    try:
        makepkg(dest_dir, noconfirm=noconfirm)
    except CalledProcessError:
        log.exception("Error building package {}".format(pkg.pkgname))
        sys.exit(EXIT_ERROR)

    # Install
    if not skip_install:
        try:
            makepkg_install(dest_dir, noconfirm=noconfirm)
        except CalledProcessError:
            log.exception("Error installing {}".format(pkg.pkgname))
            sys.exit(EXIT_ERROR)

    installed.append(pkg.pkgname)

    return installed



""" setuptools installation configuration """
import setuptools
import sink

setuptools.setup(
    name='sink',
    version=sink.__version__,
    description='Arch Linux AUR installer for the lazy',
    long_description=open('README.md').read().strip(),
    long_description_content_type='text/markdown',
    author=sink.__author__,
    author_email=sink.__email__,
    url='http://github.com/mikeshultz/sink',
    packages=['sink'],
    data_files=['README.md'],
    install_requires=[
        'coloredlogs>=10.0',
        'requests>=2.20.0',
        'semantic-version>=2.6.0'
    ],
    extras_require={
        'dev': ['wheel>=0.32.3', 'twine>=1.12.1'],
    },
    license=sink.__license__,
    zip_safe=False,
    keywords='archlinux aur installer',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',
    ],
    entry_points={
        'console_scripts': ['sink=sink.cli:main'],
    }
)
